import cProfile
import sys
from pstats import Stats
from typing import Annotated, Optional

import bugsnag
import typer
from rich.console import Group
from rich.panel import Panel
from rich.text import Text

from cerebrium import api, __version__
from cerebrium.api import cerebrium_request
from cerebrium.config import (
    get_validated_config,
)
from cerebrium.utils.check_cli_version import print_update_cli_message
from cerebrium.utils.deploy import package_app, get_function_names
from cerebrium.utils.display import confirm_partner_service_deployment
from cerebrium.utils.logging import cerebrium_log, console
from cerebrium.utils.project import get_current_project_context

cortex_cli = typer.Typer(no_args_is_help=True)


@cortex_cli.command("deploy")
def deploy(
    name: Annotated[
        Optional[str],
        typer.Option(
            "--name",
            help="Name of the App. Overrides the value in the TOML file if provided.",
        ),
    ] = None,
    disable_syntax_check: Annotated[
        bool, typer.Option(help="Flag to disable syntax check.")
    ] = False,
    log_level: Annotated[
        str,
        typer.Option(
            help="Log level for the Cortex deployment. Can be one of 'DEBUG' or 'INFO'",
        ),
    ] = "INFO",
    config_file: Annotated[
        str,
        typer.Option(
            help="Path to the cerebrium config TOML file. You can generate a config using `cerebrium init`."
        ),
    ] = "./cerebrium.toml",
    disable_confirmation: Annotated[
        bool,
        typer.Option(
            "--disable-confirmation",
            "-y",
            help="Disable the confirmation prompt before deploying.",
        ),
    ] = False,
    disable_animation: Annotated[
        bool,
        typer.Option(
            "--disable-animation",
            help="Disable TQDM loading bars and yaspin animations.",
        ),
    ] = False,
    disable_build_logs: Annotated[
        bool,
        typer.Option("--disable-build-logs", help="Disable build logs during a deployment."),
    ] = False,
):
    """
    Deploy a new Cortex app to Cerebrium. Run `cerebrium deploy --help` for more information.\n
        \n
    Usage: cerebrium deploy [OPTIONS]\n
    \n
      Deploy a Cortex app to Cerebrium.\n
    \n
    Options:\n
      --name TEXT                    Name of the App. Overrides the value in the TOML file if provided.\n
      --disable-syntax-check         Flag to disable syntax check.\n
      --log-level [DEBUG|INFO]       Log level for the Cortex deployment. Can be one of 'DEBUG' or 'INFO'.\n
      --config-file PATH             Path to the cerebrium config TOML file. You can generate a config using `cerebrium init`.\n
      -y, --disable-confirmation     Disable the confirmation prompt before deploying.\n
      --disable-animation            Disable TQDM loading bars and yaspin animations.\n
      --disable-build-logs           Disable build logs during a deployment.\n
      -h, --help                     Show this message and exit.\n
    \n
    Examples:\n
      # Deploy an app with the default settings\n
      cerebrium deploy\n
    \n
      # Deploy an app with a custom name and disabled syntax check\n
      cerebrium deploy --name my_app --disable-syntax-check\n
    """
    print_update_cli_message()

    log_level = log_level.upper()
    assert log_level in [
        "DEBUG",
        "INFO",
        "WARNING",
        "ERROR",
        "INTERNAL",
    ], "Log level must be one of 'DEBUG' or 'INFO'"

    with cProfile.Profile() as pr:
        project_id = get_current_project_context()
        config = get_validated_config(config_file, name, False)

        if config.partner_services is not None:
            payload = config.to_payload()
            if not disable_confirmation and not confirm_partner_service_deployment(config):
                sys.exit()

            console.print("Deploying partner service app...")
            payload["cliVersion"] = __version__
            setup_response = cerebrium_request(
                "POST",
                f"v2/projects/{project_id}/partner-apps",
                payload,
                requires_auth=True,
            )

            if setup_response is None or setup_response.status_code != 200:
                error_message = "Error deploying partner app. Please check the dashboard or contact support if the issue persists."
                try:
                    if setup_response is not None:
                        response_json = setup_response.json()
                        if isinstance(response_json, dict) and isinstance(
                            response_json.get("message"), str
                        ):
                            error_message = response_json["message"]
                except Exception:
                    pass

                console.print(Text(error_message, style="red"))
                bugsnag.notify(Exception(error_message))
                raise typer.Exit(1)

            dashboard_url = f"{api.dashboard_url}/projects/{project_id}/apps/{project_id}-{config.deployment.name}"
            app_endpoint = setup_response.json()["internalEndpoint"]
            info_string = f"App Dashboard: {dashboard_url}\n\nEndpoint:\n[bold red]POST[/bold red] {app_endpoint}"
            dashboard_info = Panel(
                info_string,
                title=f"[bold green] {config.deployment.name} is now live!  ",
                border_style="green",
                width=140,
            )
            console.print(Group(dashboard_info))
            return typer.Exit(0)

        build_status, setup_response = package_app(
            config,
            disable_build_logs,
            log_level,  # type: ignore
            disable_syntax_check,
            disable_animation,
            disable_confirmation,
        )
        if setup_response is None:
            message = "Error building app. Please check the dashboard or contact support if the issue persists."
            cerebrium_log(message=message, color="red")
            bugsnag.notify(Exception(message))
            raise typer.Exit(1)
        if build_status == "success":
            project_id = setup_response["projectId"]
            endpoint = setup_response["internalEndpoint"]
            dashboard_url = f"{api.dashboard_url}/projects/{project_id}/apps/{project_id}-{config.deployment.name}"
            info_string = f"App Dashboard: {dashboard_url}\n\nEndpoints:"
            function_names = get_function_names(config.custom_runtime is not None)
            for method, function_name in function_names:
                info_string += f"\n[bold red]{method}[/bold red] {endpoint}/" + function_name

            dashboard_info = Panel(
                info_string,
                title=f"[bold green] {config.deployment.name} is now live!  ",
                border_style="green",
                width=140,
            )
            console.print(Group(dashboard_info))
        elif build_status in ["build_failure", "init_failure"]:
            bugsnag.notify(Exception("User build failed"))
            console.print(Text("Unfortunately there was an issue with your build", style="red"))
            raise typer.Exit(1)
    pr.disable()
    if log_level == "INTERNAL":
        Stats(pr).sort_stats("tottime").print_stats(10)
